@@ -389,7 +389,7 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 		} else {
 			int i;
 			for(i = 0; i <= 1; i++) {
-				if(fingerInfos[i].down) {
+				if(fingerInfos[i].id != -1) {
 					currentCenterX = fingerInfos[i].x;
 					currentCenterY = fingerInfos[i].y;
 				}
@@ -438,7 +438,7 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 						/* As we only support one interval, only use larger axis. */
 						if(lastScrollXIntv < lastScrollYIntv) {
 							dirY = 0;
-						} else if(lastScrollYIntv < lastScrollXIntv) {
+						} else {
 							dirX = 0;
 						}
 					}
@@ -451,6 +451,7 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 						/* We will never reach this, but removes warning */
 						intv = 100000;
 					}
+					if(inDebugMode()) printf("Really start easing\n");
 					startEasing(currentProfile, dirX, dirY, intv);
 				}
 			}
@@ -482,7 +483,7 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 		/* Fake single-touch move event */
 		int i;
 		for(i = 0; i <= 1; i++) {
-			if(fingerInfos[i].down) {
+			if(fingerInfos[i].id != -1) {
 				movePointer(fingerInfos[i].x, fingerInfos[i].y);
 			}
 		}
@@ -499,7 +500,7 @@ void processFingerGesture(FingerInfo* fingerInfos, int fingersDown, int fingersW
 			/* Fake single-touch move event */
 			int i;
 			for(i = 0; i <= 1; i++) {
-				if(fingerInfos[i].down) {
+				if(fingerInfos[i].id != -1) {
 					movePointer(fingerInfos[i].x, fingerInfos[i].y);
 				}
 			}
