@@ -84,8 +84,8 @@ static void daemonize(void) {
 }
 
 /* Finger information */
-FingerInfo fingerInfos[2] = { { .x=0, .y=0, .down = 0, .id = 0 }, { .x=0, .y=0,
-		.down = 0, .id = 1 } };
+FingerInfo fingerInfos[2] = { { .rawX=0, .rawY=0, .id = -1, .setThisTime = 0 }, { .rawX=0, .rawY=0,
+		.id = -1, .setThisTime = 0 } };
 
 /* X stuff */
 Display* display;
@@ -99,8 +99,9 @@ Time fairlyCurrentTime;
 
 /* Calibration data */
 int calibMinX, calibMaxX, calibMinY, calibMaxY;
-double calibFactorX, calibFactorY;
 unsigned char calibSwapX, calibSwapY, calibSwapAxes;
+float calibMatrix[6];
+int calibMatrixUse = 0;
 
 /* The width and height of the screen in pixels */
 unsigned int screenWidth, screenHeight;
@@ -117,6 +118,7 @@ int fingersWereDown = 0;
 int buttonDown = 0;
 
 
+
 /* Handle errors by, well, throwing them away. */
 int invalidWindowHandler(Display *dsp, XErrorEvent *err) {
 	return 0;
@@ -387,8 +389,56 @@ Window getCurrentWindow() {
 	}
 }
 
+
+/* Sets the calibrated x, y coordinates from the raw coordinates in the given FingerInfo */
+void calibrate(FingerInfo* fingerInfo) {
+	float xf; float yf;
+	if (calibSwapAxes) {
+
+		xf = ((float)(fingerInfo->rawY - calibMinX))/((float) (calibMaxX-calibMinX));
+		yf = ((float)(fingerInfo->rawX - calibMinY))/((float) (calibMaxY-calibMinY));
+
+	} else {
+
+		xf = ((float)(fingerInfo->rawX - calibMinX))/((float) (calibMaxX-calibMinX));
+		yf = ((float)(fingerInfo->rawY - calibMinY))/((float) (calibMaxY-calibMinY));
+
+	}
+	if (calibSwapX) xf = 1 - xf;
+	if (calibSwapY) yf = 1 - yf;
+
+	if(calibMatrixUse) {
+		float xfold = xf;
+		/* Apply matrix transformation */
+		xf = xf * calibMatrix[0] + yf * calibMatrix[1] + calibMatrix[2];
+		yf = xfold * calibMatrix[3] + yf * calibMatrix[4] + calibMatrix[5];
+	}
+
+	fingerInfo->x = xf * screenWidth;
+	fingerInfo->y = yf * screenHeight;
+
+	if (fingerInfo->x < 0)
+		fingerInfo->x = 0;
+	if (fingerInfo->y < 0)
+		fingerInfo->y = 0;
+	if (fingerInfo->x > screenWidth)
+		fingerInfo->x = screenWidth;
+	if (fingerInfo->y > screenHeight)
+		fingerInfo->y = screenHeight;
+
+
+}
+
 /* Process the finger data gathered from the last set of events */
 void processFingers() {
+	int i;
+	fingersDown = 0;
+	for(i = 0; i < 2; i++) {
+		if(fingerInfos[i].id != -1) {
+			calibrate(&(fingerInfos[i]));
+			fingersDown++;
+		}
+	}
 
 	/* If we should activate at release, do it. */
 	if (buttonDown == 0 && fingersDown == 0 && activateAtRelease != 0) {
@@ -492,13 +542,6 @@ void windowMapped(Window w) {
 void setScreenSize(XRRScreenChangeNotifyEvent * evt) {
 	screenWidth = evt->width;
 	screenHeight = evt->height;
-	if(calibMaxX != calibMinX && calibMaxY != calibMinY) {
-		calibFactorX = ( screenWidth / (double) (calibMaxX - calibMinX));
-		calibFactorY = ( screenHeight / (double) (calibMaxY - calibMinY));
-	} else {
-		calibFactorX = 1.;
-		calibFactorY = 1.;
-	}
 	if(debugMode) {
 		printf("New screen size: %i x %i\n", screenWidth, screenHeight);
 	}
@@ -639,14 +682,34 @@ void readCalibrationData(int exitOnFail) {
 		calibMinY = data[2];
 		calibMaxY = data[3];
 	}
-	calibFactorX = ( screenWidth / (double) (calibMaxX - calibMinX));
-	calibFactorY = ( screenHeight / (double) (calibMaxY - calibMinY));
-	printf("Calibration factors: %f %f\n", calibFactorX, calibFactorY);
 
 	if(data != NULL) {
 		XFree(data);
 	}
 
+	float * data4 = NULL;
+	if(XIGetProperty(display, deviceID, XInternAtom(display,
+			"Coordinate Transformation Matrix", 0), 0, 9 * 32, False, XInternAtom(display,
+			"FLOAT", 0),
+			&retType, &retFormat, &retItems, &retBytesAfter,
+			(unsigned char **) &data4) != Success) {
+		data4 = NULL;
+	}
+	calibMatrixUse = 0;
+	if(data4 != NULL && retItems == 9) {
+		int i;
+		for(i = 0; i < 6; i++) {
+			/* We only take the first two rows of the matrix, the rest is unimportant anyway */
+			calibMatrix[i] = data4[i];
+		}
+		calibMatrixUse = 1;
+	}
+	if(data4 != NULL) {
+		XFree(data4);
+	}
+
+
+
 	unsigned char* data2;
 
 	if(XIGetProperty(display, deviceID, XInternAtom(display,
@@ -690,6 +753,7 @@ void readCalibrationData(int exitOnFail) {
 
 }
 
+
 /* Main function, contains kernel driver event loop */
 int main(int argc, char **argv) {
 
@@ -894,6 +958,13 @@ int main(int argc, char **argv) {
 					xLoopThreadFunction, NULL);
 
 		printf("Reading input from device ... (interrupt to exit)\n");
+
+		/* If set, twofing assumes that the device uses the multitouch slot protocol.
+		   Set to 1 the first time a ABS_MT_SLOT event occurs. Set to 0 the first time a MT_SYNC event occurs.*/
+		int useSlots = 1;
+		int currentSlot = 0;
+		/* If we don't use the slot protocol, we collect all data of one finger into tempFingerInfo and set
+		   it to the correct slot once MT_SYNC occurs. */
 		FingerInfo tempFingerInfo = { -1, -1, -1, -1 };
 
 		/* Kernel device event loop */
@@ -908,73 +979,106 @@ int main(int argc, char **argv) {
 
 					if (2 == ev[i].code) { // MT_Sync : Multitouch event end
 						/* Finger info for one finger collected in tempFingerInfo, so save it to fingerInfos. */
-						/* Currently, we assume that the tracking ids of the fingers are 0,1 -- we should not do this but rather go through all fingers and check for the tracking id. We should also check if there are more than two fingers and set all to down then. */
-						int index = tempFingerInfo.id;
-						if (index == 0 || index == 1) {
-							fingerInfos[index].id = tempFingerInfo.id;
 
-							/* Calibrate coordinates */
-							if (calibSwapAxes) {
+						if(useSlots) {
+							/* This messsage indicates we don't use the slot protocol. So
+							   set useSlots and ignore data for now. */
+							useSlots = 0;
+							currentSlot = -1;
+							if(debugMode) printf("Switch to non-slot protocol.\n");
+						} else {
 
-								fingerInfos[index].x = (int)
-										((tempFingerInfo.y - calibMinX) * calibFactorX);
-								fingerInfos[index].y = (int)
-										((tempFingerInfo.x - calibMinY) * calibFactorY);
-
-							} else {
-
-								fingerInfos[index].x = (int)
-										((tempFingerInfo.x - calibMinX) * calibFactorX);
-								fingerInfos[index].y = (int)
-										((tempFingerInfo.y - calibMinY) * calibFactorY);
-
-//								fingerInfos[index].x = ((tempFingerInfo.x
-//										- calibMinX) * displayWidth)
-//										/ (calibMaxX - calibMinX);
-//								fingerInfos[index].y = ((tempFingerInfo.y
-//										- calibMinY) * displayHeight)
-//										/ (calibMaxY - calibMinY);
+							/* Look for slot to put the data into by looking at the tracking ids */
+							int index = -1;
+							int i;
+							for(i = 0; i < 2; i++) {
+								if(fingerInfos[i].id == tempFingerInfo.id) {
+									index = i;
+									break;
+								}
+							}
+							
+							if(index == -1) {
+								for(i = 0; i < 2; i++) {
+									if(fingerInfos[i].id == -1) {
+										/* "Empty" slot, so we can add it. */
+										index = i;
+										fingerInfos[i].id = tempFingerInfo.id;
+										break;
+									}
+								}
 							}
-							if (calibSwapX)
-								fingerInfos[index].x = screenWidth
-										- fingerInfos[index].x;
-							if (calibSwapY)
-								fingerInfos[index].y = screenHeight
-										- fingerInfos[index].y;
 
-							if (fingerInfos[index].x < 0)
-								fingerInfos[index].x = 0;
-							if (fingerInfos[index].y < 0)
-								fingerInfos[index].y = 0;
-							if (fingerInfos[index].x > screenWidth)
-								fingerInfos[index].x = screenWidth;
-							if (fingerInfos[index].y > screenHeight)
-								fingerInfos[index].y = screenHeight;
+							if(index != -1) {
+								fingerInfos[index].setThisTime = 1;
+								fingerInfos[index].rawX = tempFingerInfo.rawX;
+								fingerInfos[index].rawY = tempFingerInfo.rawY;
+							}
 						}
-						fingerInfos[index].down = 1;
 					} else if (0 == ev[i].code) { // Ev_Sync event end
+
+						if(!useSlots) {
+							/* Clear slots not set this time */
+							int i;
+							for(i = 0; i < 2; i++) {
+								if(fingerInfos[i].setThisTime) {
+									fingerInfos[i].setThisTime = 0;
+								} else {
+									/* Clear slot */
+									fingerInfos[i].id = -1;
+								}
+							}
+						}
+
+
 						/* All finger data received, so process now. */
 						processFingers();
 
-						/* Reset for next set of events */
-						fingerInfos[0].down = 0;
-						fingerInfos[1].down = 0;
-						fingersDown = 0;
-						tempFingerInfo.id = -1;
+						if(!useSlots) {
+							/* If we don't use the slot protocol, reset for next set of events */
+							tempFingerInfo.id = -1;
+						}
 					}
 				} else if (ev[i].type == EV_MSC && (ev[i].code == MSC_RAW
 						|| ev[i].code == MSC_SCAN)) {
+
+				} else if (ev[i].code == 47) {
+					/* Slot event */
+					if(!useSlots) {
+						useSlots = 1;
+						if(debugMode) printf("Switch to slots protocol.\n");
+					}
+					currentSlot = ev[i].value;
+					if(currentSlot < 0 || currentSlot > 1) currentSlot = -1;
 				} else {
 					/* Set finger info values for current finger */
 					if (ev[i].code == 57) {
-						tempFingerInfo.id = ev[i].value;
+						/* ABS_MT_TRACKING_ID */
+						if(useSlots) {
+							if(currentSlot != -1) {
+								fingerInfos[currentSlot].id = ev[i].value;
+							}
+						} else {
+							tempFingerInfo.id = ev[i].value;
+						}
 					};
 					if (ev[i].code == 53) {
-						tempFingerInfo.x = ev[i].value;
-						fingersDown++;
+						if(useSlots) {
+							if(currentSlot != -1) {
+								fingerInfos[currentSlot].rawX = ev[i].value;
+							}
+						} else {
+							tempFingerInfo.rawX = ev[i].value;
+						}
 					};
 					if (ev[i].code == 54) {
-						tempFingerInfo.y = ev[i].value;
+						if(useSlots) {
+							if(currentSlot != -1) {
+								fingerInfos[currentSlot].rawY = ev[i].value;
+							}
+						} else {
+							tempFingerInfo.rawY = ev[i].value;
+						}
 					};
 				}
 			}
